program SplineDemo;
{****************************************************************************
** Demonstration of drawing a Catmull-Rom spline, a curved line that       **
** passes through a number of control points.                              **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}

uses Crt;

const
  NumPts = 5;
  Resolution = 10;

type
  BufferType = Array [0..63999] of Byte;

  PointType = record
    x, y : Real;
  end;

var
  {The screen itself}
  Screen  : BufferType absolute $A000:0000;
  {An offscreen buffer}
  Buffer  : ^BufferType;
  {Catmull-rom coefficients}
  A, B, C, D : PointType;
  {Control points}
  P : Array [-1..NumPts + 1] of PointType;

{Switches to VGA 320x200 256 colour}
procedure VGAScreen; assembler;
asm
  mov ax, $13
  int $10
end;

{Switches to text 80x25}
procedure TextScreen; assembler;
asm
  mov ax, $3
  int $10
end;

{Sets a pixel in the buffer to a specific colour}
procedure PutPixel (x, y : Word; c : Byte);
begin
  if (x < 320) and (y < 200) then
    Buffer^ [y shl 8 + y shl 6 + x] := c;
end;

{Draws a line between 2 points}
procedure DrawLine (x1, y1, x2, y2 : Integer; c : Byte);
var
  DeltaX, DeltaY, NumPixels, Counter,
  D, Dinc1, Dinc2,
  X, Xinc1, Xinc2,
  Y, Yinc1, Yinc2 : Integer;

begin
  {Calculate difference in x and y}
  DeltaX := abs (x2 - x1);
  DeltaY := abs (y2 - y1);

  {Check accuracy. This determines the independent variable}
  if (DeltaX >= DeltaY) then begin
    {X is the independent variable}
    NumPixels := Deltax + 1;
    D := (DeltaY shl 1) - DeltaX;
    Dinc1 := DeltaY shl 1;
    Dinc2 := (DeltaY - DeltaX) shl 1;
    Xinc1 := 1;
    Xinc2 := 1;
    Yinc1 := 0;
    Yinc2 := 1;
  end else begin
    {Y is the independent variable}
    NumPixels := DeltaY + 1;
    D := (DeltaX shl 1) - DeltaY;
    Dinc1 := DeltaX shl 1;
    Dinc2 := (DeltaX - DeltaY) shl 1;
    Xinc1 := 0;
    Xinc2 := 1;
    Yinc1 := 1;
    Yinc2 := 1;
  end;

  {Make sure x and y move in the right directions}
  if x1 > x2 then begin
    Xinc1 := -Xinc1;
    Xinc2 := -Xinc2;
  end;
  if y1 > y2 then begin
    Yinc1 := -Yinc1;
    Yinc2 := -Yinc2;
  end;

  {Start drawing at x1, y1}
  X := x1;
  Y := y1;

  {Draw the pixels}
  for Counter := 1 to NumPixels do begin
    {Plot a pixel}
    PutPixel (X, Y, c);
    {Update coordinates}
    if (D < 0) then begin
      inc (D, Dinc1);
      inc (X, Xinc1);
      inc (Y, Yinc1);
    end else begin
      inc (D, Dinc2);
      inc (X, Xinc2);
      inc (Y, Yinc2);
    end;
  end;
end;

procedure SplinePoint (t : Real; var Point : PointType);
var
  t2, t3 : Real;

begin
  {Square and cube of t}
  t2 := t * t;
  t3 := t2 * t;
  {Calculate coordinates}
  Point.x := ((A.x * t3) + (B.x * t2) + (C.x * t) + D.x) / 2;
  Point.y := ((A.y * t3) + (B.y * t2) + (C.y * t) + D.y) / 2;
end;

{Computes coefficients for point n.
This is a matrix transform:
  -1  3 -3  1
   2 -5  4 -1
  -1  0  1  0
   0  2  0  0
}
procedure ComputeCoeffs (n : Integer);
begin
  {x-coefficients}
  A.x :=    -P [n - 1].x + 3 * P [n].x - 3 * P [n + 1].x + P [n + 2].x;
  B.x := 2 * P [n - 1].x - 5 * P [n].x + 4 * P [n + 1].x - P [n + 2].x;
  C.x :=    -P [n - 1].x                   + P [n + 1].x;
  D.x :=                   2 * P [n].x;

  {y-coefficients}
  A.y :=    -P [n - 1].y + 3 * P [n].y - 3 * P [n + 1].y + P [n + 2].y;
  B.y := 2 * P [n - 1].y - 5 * P [n].y + 4 * P [n + 1].y - P [n + 2].y;
  C.y :=    -P [n - 1].y                   + P [n + 1].y;
  D.y :=                   2 * P [n].y;
end;

procedure DrawSpline (Points : Integer; Colour : Byte);
var
  Point, Segment : Integer;
  Current, Next : PointType;

begin
  P [0]          := P [1];
  P [Points + 1] := P [Points];

  {Loop along all the points, drawing a line to the next point}
  for Point := 1 To Points - 1 do begin
    {Calculate coefficients for this point}
    ComputeCoeffs (Point);
    {Calculate the start point for the first segment}
    SplinePoint (0, Current);
    {Split into smaller segments}
    for Segment := 1 To Resolution Do Begin
      {Calculate end point}
      SplinePoint (Segment / Resolution, Next);
      {Draw segment}
      DrawLine (Round (Current.x),
                Round (Current.y),
                Round (Next.x),
                Round (Next.y),
                Colour);
      {Next part}
      Current := Next;
    end;
  end;
end;

var
  Point : Byte;

begin
  Randomize;

  {Switch to VGA screen}
  VGAScreen;

  {I'm drawing to the screen instead of a buffer}
  Buffer := @Screen;

  repeat
    {Clear screen}
    FillChar (Mem [$A000:0000], 64000, 0);

    {Get random points}
    for Point := 1 to NumPts do begin
      P [Point].x := Random (320);
      P [Point].y := Random (200);
    end;

    {Draw the spline along those points}
    DrawSpline (NumPts, 8);

    {Draw the points themselves}
    for Point := 1 to NumPts do
      PutPixel (Round (P [Point].x), Round (P [Point].y), 7);

  until ReadKey = Chr (27);

  TextScreen;
end.